#version 330 core

// Uniforms which are always available.
uniform mat4 u_worldTrans;
uniform vec3 u_cameraPos;
uniform float u_bodySize;
uniform vec2 u_cameraNearFar;
uniform float u_cameraK;
uniform float u_time;
uniform float u_simuTime;
// Unit conversion.
uniform float u_kmToU;
// Aurora footprints are in the diffuse cubemap.
uniform samplerCube u_diffuseCubemap;
// The SDF is in the specular cubmeap.
uniform samplerCube u_specularCubemap;

#include <shader/lib/cubemap.glsl>

#define fetchTexFootprint(texCoord) texture(u_diffuseCubemap, UVtoXYZ(texCoord))
#define fetchTexSdf(texCoord) texture(u_specularCubemap, UVtoXYZ(texCoord))

// We use the diffuse channel for the main aurora color.
#ifdef diffuseColorFlag
uniform vec4 u_diffuseColor;
#define fetchColorDiffuse(defaultValue) u_diffuseColor
#else
#define fetchColorDiffuse(defaultValue) defaultValue
#endif // diffuseColorFlag

// We use the emissive channel for the top part.
#ifdef emissiveColorFlag
uniform vec4 u_emissiveColor;
#define fetchColorEmissive(defaultValue) u_emissiveColor
#else
#define fetchColorEmissive(defaultValue) defaultValue
#endif // diffuseColorFlag

#ifdef ssrFlag
#include <shader/lib/ssr.frag.glsl>
#endif // ssrFlag

// INPUT
struct VertexData {
    vec2 texCoords;
    vec3 normal;
    vec3 viewDir;
    vec3 ambientLight;
    float opacity;
    vec4 color;
    vec3 fragPosWorld;
    mat3 tbn;
};
in VertexData v_data;

// OUTPUT
layout(location = 0) out vec4 fragColor;
layout(location = 1) out vec4 layerBuffer;

#include <shader/lib/logdepthbuff.glsl>

#ifndef PI
#define PI 3.14159265
#endif // PI
#define PI2 2.0 * 3.14159265
#define PI_2 3.14159265 / 2.0
#define PI_3 3.14159265 / 3.0
#define PI_5 3.14159265/ 5.0

#define AURORA_RADIUS 6650.0
#define ATM_RADIUS 6550.0
#define ATM_RADIUS98 (ATM_RADIUS * 0.975)
#define AURORA_MINUS_ATM (AURORA_RADIUS - ATM_RADIUS98)
#define EARTH_RADIUS (6371.1 * u_kmToU)
#define U30 (30.0 * u_kmToU)
#define BASE_STEP (25.0 * u_kmToU)

// This is the color mapped to the lowest point
// of the aurora. The main and high colors come in the diffuse and emissive color
// channels.
const vec3 lowColor = vec3(0.35, 2.4, 2.51);

float hash(float h) {
    return fract(sin(h) * 43758.5453123);
}

float hash31(vec3 p3) {
    p3 = fract(p3 * .1031);
    p3 += dot(p3, p3.zyx + 31.32);
    return fract((p3.x + p3.y) * p3.z);
}

float noise(vec3 x) {
    vec3 p = floor(x);
    vec3 f = fract(x);
    f = f * f * (3.0 - 2.0 * f);

    float n = p.x + p.y * 157.0 + 113.0 * p.z;
    return mix(
        mix(mix(hash(n + 0.0), hash(n + 1.0), f.x),
            mix(hash(n + 157.0), hash(n + 158.0), f.x), f.y),
        mix(mix(hash(n + 113.0), hash(n + 114.0), f.x),
            mix(hash(n + 270.0), hash(n + 271.0), f.x), f.y), f.z);
}
float fbm(vec3 pos) {
    float f;
    f = 0.7000 * noise(pos);
    pos *= 2.01;
    f = 0.5000 * noise(pos);
    pos *= 2.01;
    f += 0.2500 * noise(pos);
    pos *= 2.02;
    f += 0.1250 * noise(pos);

    return f;
}

struct Intersection {
    float t0;
    float t1;
    int hits; // forward hits
    bool isInside;
};
Intersection sphereIntersection(vec3 origin, vec3 ray, float radius) {
    float b = -dot(origin, ray);
    vec3 w = origin + b * ray;
    float r2 = radius * radius;
    float d = r2 - dot(w, w);
    if (d < 0.0) {
        return Intersection(0.0, 0.0, 0, false);
    }
    if (d == 0.0) {
        return Intersection(b, b, 1, false);
    }
    float c = dot(origin, origin) - r2;
    float sqrtd = sqrt(d);
    float q = b + sign(b) * sqrtd;
    float t0 = c / q;
    float t1 = q;
    int hits = int(t0 >= 0.0) + int(t1 > 0.0);
    return Intersection(min(t0, t1), max(t0, t1), hits, c < 0.0);
}

float angle(vec3 a, vec3 b) {
    return acos(clamp(dot(normalize(a), normalize(b)), -1.0, 1.0));
}

vec2 uv(vec3 p, vec3 rayDir, vec3 camPos) {
    if (dot(rayDir, p - camPos) > 0.000) {
        // Ok.
        p = normalize(p);
        return vec2(0.5 - atan(-p.z, -p.x) / (PI2), asin(-p.y) / PI + 0.5);
    } else {
        // Invalid result.
        return vec2(5.0, 5.0);
    }
}

vec3 auroraColor(float h, vec3 main, vec3 high) {
    if (h < 0.2) {
        return mix(lowColor, main, h / 0.2);
    } else {
        return mix(main, high, pow(h, 6.5));
    }
}
vec3 ACESFilm(vec3 x) {
    return clamp((x * (2.51 * x + 0.03)) / (x * (2.43 * x + 0.59) + 0.14), 0.0, 1.0);
}

vec4 raymarchAurora(vec3 camPos, vec3 rayDir, vec3 objPos, float radius, vec3 mainColor, vec3 highColor) {
    // Intersection distances with aurora sphere.
    Intersection iaurora = sphereIntersection(camPos, rayDir, radius);
    // Intersection distances with the atmosphere.
    // We do this to have a full length that always works.
    Intersection iearth = sphereIntersection(camPos, rayDir, EARTH_RADIUS);
    if (iearth.hits > 0) {
        iaurora.t1 = iearth.t0;
    }
    vec3 p;
    // Apply some jitter to the initial depth to remove the banding.
    float jitter = hash31(rayDir * 5000.0 + u_time) * BASE_STEP;
    float depth = iaurora.t0 + jitter;
    float d = -1.0;

    vec3 sum = vec3(0.0, 0.0, 0.0);
    while (depth < iaurora.t1) {
        p = camPos + rayDir * depth;
        float heightKm = length(p) / u_kmToU;

        vec2 tc = uv(p, rayDir, camPos);
        float t = mod(u_simuTime * 0.0001, PI2);
        float inner = 50.0 * (PI_3 + tc.y + t);
        vec2 tcwarp = vec2(sin(inner) * 0.01, cos(2.0 * inner + PI_5) * 0.004);
        tc += tcwarp;
        float dx = (1.0 - fetchTexSdf(tc).g) * BASE_STEP;
        depth += dx;
        if (dx < U30 && heightKm > ATM_RADIUS98) {
            // Sample footprint at point.
            float footprint = fetchTexFootprint(tc).g;
            if (footprint > 0.001) {
                float scaleHeightKm = heightKm - ATM_RADIUS98;
                // Deposition factor, decreases with scaleHeight.
                float deposition = pow(1.0 - scaleHeightKm / AURORA_MINUS_ATM, 4.0);
                float depositionFactor = deposition * smoothstep(1.0, 0.4, deposition);

                // The factor uses FBM'd noise.
                float fac = pow(fbm(vec3(tc, 1.0) * 290.0 + u_time * 0.2), 2.0);
                // Uncomment next line to add some large-scale noise to turn on and off
                // entire parts of the aurora.
                // fac *= clamp(fbm(vec3(tc, 0.5) * 2.0 + u_simuTime * 0.001) * 2.0 - 0.5, 0.0, 1.0);
                // Factor to modulate aurora intensity.
                float modulation = fac * depositionFactor;

                sum += footprint * auroraColor(1.0 - deposition, mainColor, highColor) * modulation;
                if (footprint > 0.01 && d < 0.0) {
                    d = depth;
                }
            }
        }
    }
    return vec4(sum, d);
}

void main() {
    // Directions and positions.
    // Object is always at origin in our reference frame.
    vec3 camPos = u_cameraPos;
    vec3 rayDir = normalize(v_data.fragPosWorld);
    vec3 objPos = vec3(0.0);
    float radius = u_bodySize * 0.5;

    // Colors.
    vec4 mainColor = fetchColorDiffuse(vec4(0.0, 1.0, 0.0, 1.0));
    vec4 highColor = fetchColorEmissive(vec4(1.0, 0.01, 0.4, 1.0));

    vec4 sum = raymarchAurora(camPos, rayDir, objPos, radius, mainColor.rgb, highColor.rgb);

    // Tonemapping
    sum.rgb = ACESFilm(sum.rgb);

    // Gamma correction 1.0/2.2 = 0.4545...
    sum.rgb = pow(sum.rgb, vec3(0.4545));

    fragColor = vec4(sum.rgb * mainColor.a, 1.0);

    // Logarithmic depth buffer.
    if (sum.a > 0.0) {
        gl_FragDepth = getDepthValue(sum.a * 0.8, u_cameraNearFar.y, u_cameraK);
    }

    #ifdef ssrFlag
    ssrBuffers();
    #endif // ssrFlag
}
